/*****************************************************
 * Rainbow Cipher header file for ANSI C             *
 *   Chang-Hyi Lee  and  Jeong-Soo Kim               *
 *   Digital Communication Lab.,                     *
 *   SAIT, Samsung Advanced Institute of Technology  *
 *   Updated and edited by Alexander Myasnikov       *
 *   Web: www.darksoftware.narod.ru                  *
 *****************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*  Defines:  */
#define     BITSPERBLOCK    128	/*  Number of bits in a cipher block  */
#define     BLOCKSIZE       (BITSPERBLOCK/8)	/* # bytes in a cipher block  */
#define     BLOCK_WSIZE     (BITSPERBLOCK/32)	/* # u32's in a cipher block  */
#define     R_TRUE            1
#define     R_FALSE           0


/*  Error Codes :  */

#define     MAX_KEY_SIZE	   32	/* # of ASCII char's needed to   represent a key */
#define     MAX_IV_SIZE		   16	/* # bytes needed to represent an IV  */
#define     SCHEDULE_KEY_SIZE  16*2*(R+1)	/* total size of scheduled key */
#define     R    7		/* proposed encryption round */


typedef unsigned char u8;	/*  8 bit */
typedef unsigned short u16;	/* 16 bit */
typedef unsigned long u32;	/* 32 bit */


u8 RED[512];			/* The S-box table RED=[f]|[f^(-1)] */

/*  The structure for key information */
typedef struct keyInstance
{
  char keyMaterial[MAX_KEY_SIZE];	/*  Raw key data in ASCII */
  u8 KS_Enc[SCHEDULE_KEY_SIZE];	/*  encryption key */
  u8 KS_Dec[SCHEDULE_KEY_SIZE];	/*  decryption key */
} keyInstance;


/*  Function protoypes  */
void makeKey (keyInstance * key, char *keyMaterial);
void cipherInit (void);

#define     WD(a)    ((u32 *)(a))	/* for utility */


#	define HARDWARE_ROTATIONS
#	define ASSEMBLER_CORE



#	define ROTL(x, s) (_lrotl ((x), (s)))
#	define ROTR(x, s) (_lrotr ((x), (s)))


#		define GETB0(x) ((u8)  ((x)      ))
#		define GETB1(x) ((u8)  ((x) >>  8))
#		define GETB2(x) ((u8)  ((x) >> 16))
#		define GETB3(x) ((u8)  ((x) >> 24))

#	define PUTB0(x) ((u32) (x)      )
#	define PUTB1(x) ((u32) (x) <<  8)
#	define PUTB2(x) ((u32) (x) << 16)
#	define PUTB3(x) ((u32) (x) << 24)

#define COPY_BLOCK(trg, src) \
{ \
	(trg)[0] = (src)[0]; \
	(trg)[1] = (src)[1]; \
	(trg)[2] = (src)[2]; \
	(trg)[3] = (src)[3]; \
}				/* COPY_BLOCK */


/******** from here, represented key generation ********/

static void
rainbowMixing (u32 * roundKey)
{				/* key mixing procedure via rorarions and XOR operations */
  roundKey[0] = ROTR (roundKey[0], 3) ^
    ROTR (roundKey[1], 5) ^
    ROTR (roundKey[2], 7) ^ ROTR (roundKey[3], 11) ^ 0xb7e15163;
  roundKey[1] = ROTR (roundKey[0], 5) ^
    ROTR (roundKey[1], 7) ^
    ROTR (roundKey[2], 11) ^ ROTR (roundKey[3], 3) ^ 0xb7e15163;
  roundKey[2] = ROTR (roundKey[0], 7) ^
    ROTR (roundKey[1], 11) ^
    ROTR (roundKey[2], 3) ^ ROTR (roundKey[3], 5) ^ 0xb7e15163;
  roundKey[3] = ROTR (roundKey[0], 11) ^
    ROTR (roundKey[1], 3) ^
    ROTR (roundKey[2], 5) ^ ROTR (roundKey[3], 7) ^ 0xb7e15163;
}				/* rainbowMixing */

/* make decryption key to satisfy the cipher's self reciprocality(self-invertable) */
#define SUB_KEY_SCROL(K_, K, M, i) \
{ \
	(K_)[i] = ((K)[0]&(M)[i])^((K)[1]&(M)[(i+1)%4])^   \
		((K)[2]&(M)[(i+2)%4])^((K)[3]&(M)[(i+3)%4]); \
}

void
makeKey (keyInstance * key, char *keyMaterial)
{
  u32 *get, *encKey, *decKey;
  int i, j;

  memcpy (&key->keyMaterial, &keyMaterial, 32);

  get = WD (keyMaterial);
  encKey = WD (key->KS_Enc);
  COPY_BLOCK (encKey, get);	/* put the key material to the first key-block */

  i = 4;
  encKey += 4;
  COPY_BLOCK (encKey, (encKey - 4));

  get += 4;
  for (j = 0; j < i; j++)
    encKey[j] ^= get[j];
  /* to make it applicable to |key material|=128,160,192,...,256 in bits */
  rainbowMixing (encKey);

  for (i = 2; i < 2 * (R + 1); i++)
    {
      /* apply the rainbowMixing function: */
      encKey += 4;
      COPY_BLOCK (encKey, encKey - 4);
      rainbowMixing (encKey);
    }


  encKey = (WD (key->KS_Enc)) + 4;
  for (j = 0; j < R + 1; j++)
    {
      encKey[0] = encKey[1] ^ encKey[2] ^ encKey[3] ^ 0xffffffffL;
      encKey += 8;
    }				/* to make the key-masking be self-invertable */

  decKey = (WD (key->KS_Dec)) + 4;
  for (j = 0; j < R + 1; j++)
    {
      encKey -= 8;
      COPY_BLOCK (decKey, encKey);
      decKey += 8;
    }

  encKey = (WD (key->KS_Enc)) + 8 * R;
  decKey = WD (key->KS_Dec);

  /* to make the whole encryption process be self-reciprocal! */
  for (j = 0; j < R + 1; j++)
    {
      SUB_KEY_SCROL (decKey, encKey, (encKey + 4), 0);
      SUB_KEY_SCROL (decKey, encKey, (encKey + 4), 1);
      SUB_KEY_SCROL (decKey, encKey, (encKey + 4), 2);
      SUB_KEY_SCROL (decKey, encKey, (encKey + 4), 3);
      decKey += 8;
      encKey -= 8;
    }


}				/* rainbow-RoundKeys are generated */

typedef u8 base[8];

u8 RDC[15];			/* polynomial,p(x), reduction table */
u8 MTX[8];			/* basis changing matrix */

static u8 field_mul (u8 a, u8 b);
static void make_MTX (base b1, base b2);
static u8 coef_cnv (u8 coef, base MATRIX);



void
cipherInit ()
{
  u8 irrpoly = 0xa9, g, t1, t2, t3, t;
  base pb, nb;			/* pb: standard basis;  nb: normal basis */
  int i, j;


  /*---> start: make RDC */
  for (i = 0; i < 8; i++)
    RDC[i] = 1 << i;
  RDC[8] = irrpoly;		/* = 00011011 <- x^8 + x^4 + x^3 + x + 1 */

  for (i = 9; i < 15; i++)
    {
      if (RDC[i - 1] & 0x80)
	{
	  RDC[i] = (u8) (RDC[i - 1] << 1) ^ irrpoly;
	}
      else
	{
	  RDC[i] = RDC[i - 1] << 1;
	}
    }				/* complete: make RDC */

  /* ----> start : make basis conversion matrix, MTX */
  for (i = 0; i < 8; i++)
    pb[i] = (1 << i);
  nb[0] = 2;
  for (j = 1; j < 8; j++)
    nb[j] = field_mul (nb[j - 1], nb[j - 1]);
  make_MTX (pb, nb);
  /* complete: make basis conversion matrix, MTX */

  /* final step for generating the S-box Table RED */
  RED[0] = RED[256] = 0;
  for (g = 255; g > 0; g--)
    {
      t1 = field_mul (g, g);
      t2 = field_mul (t1, t1);
      t3 = field_mul (t2, g);
      t1 = field_mul (t2, t2);
      t2 = field_mul (t1, t1);
      t1 = field_mul (t2, t2);
      t = coef_cnv (field_mul (t1, t3), MTX);	/* coefficient changing to
						   over normal basis */
      t1 = coef_cnv (g, MTX);	/* convert g in normal basis coordinates */
      RED[t1] = t;		/* store g^(37) into RED[g] in normal basis
				   coordinates */
      RED[256 + t] = t1;	/* store its inverse = g^(193) in normal
				   basis coord. */
    }				/* The S-box, RED, has been generated */

}

/* this is the field multiplication function over GF(2^8) with
   the field defining polynomial x^8 + x^4 + x^3 + x + 1 */
static u8
field_mul (u8 a, u8 b)
{
  int cf_a, cf_b;
  u8 w = 0, cf;

  for (cf_a = 0; cf_a < 8; cf_a++)
    {
      for (cf_b = 0; cf_b < 8; cf_b++)
	{
	  cf = ((a >> cf_a) & (b >> cf_b)) & 1;
	  if (cf)
	    w ^= RDC[cf_a + cf_b];
	}
    }
  return w;
}

/* generate the base conversion matrix of basis b1 to b2 */
static void
make_MTX (base b1, base b2)
{
  unsigned short cf, i, j;
  u8 tmp[8], d;

  for (i = 0; i < 8; i++)
    {
      for (cf = 0; cf < 256; cf++)
	{
	  d = 0;
	  for (j = 0; j < 8; j++)
	    d ^= (((cf >> j) & 1) * b2[j]);
	  if (d == b1[i])
	    {
	      tmp[i] = (u8) cf;
	      cf = 257;
	    }
	}
    }

  for (i = 0; i < 8; i++)
    {
      MTX[i] = 0;
      for (j = 0; j < 8; j++)
	MTX[i] ^= (((tmp[j] >> i) & 1) << j);
    }
  return;
}

#define PROD(C, A, B, k) /* inner product routine */ \
{ \
	C = A&B;     \
	C ^= (C>>4); \
	C ^= (C>>2); \
	C ^= (C>>1); \
	C &= 1;      \
}
/* coefficient conversion of coef via basis changing mtx, MTX */
static u8
coef_cnv (u8 coef, base MATRIX)
{
  u8 t, w = 0;
  int i;

  for (i = 0; i < 8; i++)
    {
      PROD (t, MATRIX[i], coef, i);
      w |= (t << i);
    }
  return w;
}


#define     WD(a)    ((u32 *)(a))	/* for utility */


static void RB_Enc_ecb (u8 * cipherKey, u8 * input, u8 * outBuffer);



#	define ROTL(x, s) (_lrotl ((x), (s)))
#	define ROTR(x, s) (_lrotr ((x), (s)))


#		define GETB0(x) ((u8)  ((x)      ))
#		define GETB1(x) ((u8)  ((x) >>  8))
#		define GETB2(x) ((u8)  ((x) >> 16))
#		define GETB3(x) ((u8)  ((x) >> 24))
#	define PUTB0(x) ((u32) (x)      )
#	define PUTB1(x) ((u32) (x) <<  8)
#	define PUTB2(x) ((u32) (x) << 16)
#	define PUTB3(x) ((u32) (x) << 24)

#define COPY_BLOCK(trg, src) \
{ \
	(trg)[0] = (src)[0]; \
	(trg)[1] = (src)[1]; \
	(trg)[2] = (src)[2]; \
	(trg)[3] = (src)[3]; \
}				/* COPY_BLOCK */


/******** from here, represented the encryption routine ********/

#define G_function(RN) \
{ /* G-layer */ \
	data[0] ^= key[(RN)][0]; \
	data[1] ^= key[(RN)][1]; \
	data[2] ^= key[(RN)][2]; \
	data[3] ^= key[(RN)][3]; \
}

#define B_function(RN) \
{ /* B-layer */ \
	tmp[0] = (data[0] & key[(RN)][0])^\
		 (data[1] & key[(RN)][1])^    \
		 (data[2] & key[(RN)][2])^    \
		 (data[3] & key[(RN)][3]);    \
	tmp[1] = (data[0] & key[(RN)][1])^\
		  (data[1] & key[(RN)][2])^   \
		  (data[2] & key[(RN)][3])^   \
		  (data[3] & key[(RN)][0]);   \
	tmp[2] = (data[0] & key[(RN)][2])^\
		  (data[1] & key[(RN)][3])^   \
		  (data[2] & key[(RN)][0])^   \
		  (data[3] & key[(RN)][1]);   \
	tmp[3] = (data[0] & key[(RN)][3])^\
		  (data[1] & key[(RN)][0])^   \
		  (data[2] & key[(RN)][1])^   \
		  (data[3] & key[(RN)][2]);   \
}

#define R_function(TABLE) \
{ /* R-layer */ \
	data[0] = PUTB1((TABLE)[GETB0(tmp[0])])|  \
		PUTB0((TABLE)[256+GETB1(tmp[0])])|    \
		PUTB3((TABLE)[GETB2(tmp[0])])|        \
		PUTB2((TABLE)[256+GETB3(tmp[0])]);    \
	data[1] = PUTB2((TABLE)[GETB0(tmp[1])])|  \
		PUTB0((TABLE)[256+GETB2(tmp[1])])|    \
		PUTB3((TABLE)[GETB1(tmp[1])])|        \
		PUTB1((TABLE)[256+GETB3(tmp[1])]);    \
	data[2] = PUTB3((TABLE)[GETB0(tmp[2])])|  \
		PUTB0((TABLE)[256+GETB3(tmp[2])])|    \
		PUTB2((TABLE)[GETB1(tmp[2])])|        \
		PUTB1((TABLE)[256+GETB2(tmp[2])]);    \
	data[3] = PUTB2((TABLE)[GETB0(tmp[3])])|  \
		PUTB0((TABLE)[256+GETB2(tmp[3])])|    \
		PUTB3((TABLE)[GETB1(tmp[3])])|        \
		PUTB1((TABLE)[256+GETB3(tmp[3])]);    \
}

#define ROUND_function(key_num) \
{ /* one round process 'F_function' in the document */ \
	G_function(key_num);   \
	B_function(key_num+1); \
	R_function(SBox);      \
}

#define ONEBLOCK_CIPH /* here : only for the blockLen=16bytes */ \
{ /* one block encryption */ \
	ROUND_function(0);  \
	ROUND_function(2);  \
	ROUND_function(4);  \
	ROUND_function(6);  \
	ROUND_function(8);  \
	ROUND_function(10); \
	ROUND_function(12); \
	G_function(14);     \
	B_function(15);     \
	COPY_BLOCK(data,tmp);\
}


/* ECB-mode encryption */
static void
RB_Enc_ecb (u8 * cipherKey, u8 * input, u8 * outBuffer)
{
  u32 tmp[4], data[4], key[2 * (R + 1)][4];
  u32 *scan, *tar;
  u8 *SBox;
  int i, ib;

  SBox = RED;
  scan = WD (cipherKey);
  for (i = 0; i < 2 * (R + 1); i++)
    {
      COPY_BLOCK (key[i], scan);
      scan += 4;
    }


  scan = WD (input);
  tar = WD (outBuffer);


  COPY_BLOCK (data, scan);
  ONEBLOCK_CIPH;		/* encrypt */
  COPY_BLOCK (tar, data);

}

#define BLOCK_XOR(B, A) \
{ \
	B[0] ^= A[0];\
	B[1] ^= A[1];\
	B[2] ^= A[2];\
	B[3] ^= A[3];\
}

#define LSHIFT_PAST(D, b) \
{ /* shift by 1bit of and paste 1bit to cipher input data */ \
	dt = D[0]>>31;       \
	D[0] = (D[0]<<1)|b;  \
	df = (D[1]<<1)|dt;   \
	dt = D[1]>>31;       \
	D[1] = df;           \
	df = (D[2]<<1)|dt;   \
	dt = D[2]>>31;       \
	D[2] = df;           \
	D[3] = (D[3]<<1)|dt; \
}


#define     WD(a)    ((u32 *)(a))	/* for utility */


static void RB_Dec_ecb (u8 * cipherKey, u8 * input, u8 * outBuffer);


#	define ROTL(x, s) (_lrotl ((x), (s)))
#	define ROTR(x, s) (_lrotr ((x), (s)))
#		define GETB0(x) ((u8)  ((x)      ))
#		define GETB1(x) ((u8)  ((x) >>  8))
#		define GETB2(x) ((u8)  ((x) >> 16))
#		define GETB3(x) ((u8)  ((x) >> 24))


#	define PUTB1(x) ((u32) (x) <<  8)
#	define PUTB2(x) ((u32) (x) << 16)
#	define PUTB3(x) ((u32) (x) << 24)

#define COPY_BLOCK(trg, src) \
{ \
	(trg)[0] = (src)[0]; \
	(trg)[1] = (src)[1]; \
	(trg)[2] = (src)[2]; \
	(trg)[3] = (src)[3]; \
}				/* COPY_BLOCK */


/******** from here, represented the decryption routine ********/

#define G_function(RN) \
{ /* G-layer */ \
	data[0] ^= key[(RN)][0]; \
	data[1] ^= key[(RN)][1]; \
	data[2] ^= key[(RN)][2]; \
	data[3] ^= key[(RN)][3]; \
}

#define B_function(RN) \
{ /* B-layer */ \
	tmp[0] = (data[0] & key[(RN)][0])^\
		 (data[1] & key[(RN)][1])^    \
		 (data[2] & key[(RN)][2])^    \
		 (data[3] & key[(RN)][3]);    \
	tmp[1] = (data[0] & key[(RN)][1])^\
		  (data[1] & key[(RN)][2])^   \
		  (data[2] & key[(RN)][3])^   \
		  (data[3] & key[(RN)][0]);   \
	tmp[2] = (data[0] & key[(RN)][2])^\
		  (data[1] & key[(RN)][3])^   \
		  (data[2] & key[(RN)][0])^   \
		  (data[3] & key[(RN)][1]);   \
	tmp[3] = (data[0] & key[(RN)][3])^\
		  (data[1] & key[(RN)][0])^   \
		  (data[2] & key[(RN)][1])^   \
		  (data[3] & key[(RN)][2]);   \
}

//#define R_function(TABLE)
#define R_function(TABLE) \
{ /* R-layer */ \
	data[0] = PUTB1((TABLE)[GETB0(tmp[0])])|  \
		PUTB0((TABLE)[256+GETB1(tmp[0])])|    \
		PUTB3((TABLE)[GETB2(tmp[0])])|        \
		PUTB2((TABLE)[256+GETB3(tmp[0])]);    \
	data[1] = PUTB2((TABLE)[GETB0(tmp[1])])|  \
		PUTB0((TABLE)[256+GETB2(tmp[1])])|    \
		PUTB3((TABLE)[GETB1(tmp[1])])|        \
		PUTB1((TABLE)[256+GETB3(tmp[1])]);    \
	data[2] = PUTB3((TABLE)[GETB0(tmp[2])])|  \
		PUTB0((TABLE)[256+GETB3(tmp[2])])|    \
		PUTB2((TABLE)[GETB1(tmp[2])])|        \
		PUTB1((TABLE)[256+GETB2(tmp[2])]);    \
	data[3] = PUTB2((TABLE)[GETB0(tmp[3])])|  \
		PUTB0((TABLE)[256+GETB2(tmp[3])])|    \
		PUTB3((TABLE)[GETB1(tmp[3])])|        \
		PUTB1((TABLE)[256+GETB3(tmp[3])]);    \
}

/*
#define ROUND_function(key_num) \
{ \
	G_function((key_num));   \
	B_function((key_num+1)); \
	R_function(SBox);        \
}
*/
#define ONEBLOCK_CIPH /* here : only for the blockLen=16bytes */\
{ /* one block encryption */ \
	ROUND_function(0);  \
	ROUND_function(2);  \
	ROUND_function(4);  \
	ROUND_function(6);  \
	ROUND_function(8);  \
	ROUND_function(10); \
	ROUND_function(12); \
	G_function(14);     \
	B_function(15);     \
	COPY_BLOCK(data,tmp);\
}


/* ECB-mode encryption */
static void
RB_Dec_ecb (u8 * cipherKey, u8 * input, u8 * outBuffer)
{
  u32 tmp[4], data[4], key[2 * (R + 1)][4];
  u32 *scan, *tar;
  u8 *SBox;
  int i, ib;

  SBox = RED;
  scan = WD (cipherKey);
  for (i = 0; i < 2 * (R + 1); i++)
    {
      COPY_BLOCK (key[i], scan);
      scan += 4;
    }


  scan = WD (input);
  tar = WD (outBuffer);


  COPY_BLOCK (data, scan);
  ONEBLOCK_CIPH;
  COPY_BLOCK (tar, data);

}

/*
#define BLOCK_XOR(B, A) \
{ \
	B[0] ^= A[0]; \
	B[1] ^= A[1]; \
	B[2] ^= A[2]; \
	B[3] ^= A[3]; \
}
*/

#define LSHIFT_PAST(D, b) \
{ /* shift by 1bit of and paste 1bit to cipher input data */ \
	dt = D[0]>>31;       \
	D[0] = (D[0]<<1)|b;  \
	df = (D[1]<<1)|dt;   \
	dt = D[1]>>31;       \
	D[1] = df;           \
	df = (D[2]<<1)|dt;   \
	dt = D[2]>>31;       \
	D[2] = df;           \
	D[3] = (D[3]<<1)|dt; \
}


keyInstance keys;

void __stdcall __export
crypt (u8 * input, u8 * outBuffer)
{

  RB_Enc_ecb (keys.KS_Enc, &input[0], &outBuffer[0]);
}


void __stdcall __export
decrypt (u8 * input, u8 * outBuffer)
{
  RB_Dec_ecb (keys.KS_Dec, &input[0], &outBuffer[0]);
}



void __stdcall __export
setup (u8 * key)
{
  cipherInit ();
  makeKey (&keys, &key[0]);
}
